# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library of functions for defining chemical motifs."""

from typing import Any, Dict
from absl import logging  # pylint: disable=unused-import

import networkx as nx  # pylint: disable=unused-import
import numpy as np

from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import shapes

EARTH_COLOR = (225, 169, 95, 255)  # An earth color.
WHITE_COLOR = (255, 255, 255, 255)  # A white color.

STOMACH = """
xxxx********xxxx
xxxx********xxxx
xxxx********xxxx
xxxx********xxxx
xxxxxxxxxxxxxxxx
xxxxxxxxxxxxxxxx
xxxxxxxxxxxxxxxx
xxxxxxxxxxxxxxxx
xx************xx
xx************xx
xx************xx
xx************xx
xxxxxxxxxxxxxxxx
xxxxxxxxxxxxxxxx
xxxxxxxxxxxxxxxx
xxxxxxxxxxxxxxxx
"""

DIAMOND_SHAPE = """
xxxabxxx
xxaabbxx
xaaabbbx
aaaabbbb
ddddcccc
xdddcccx
xxddccxx
xxxdcxxx
"""

SQUARE_SHAPE = """
bbbbbbbb
bbbbbbbb
bbbbbbbb
bbbbbbbb
bbbbbbbb
bbbbbbbb
bbbbbbbb
bbbbbbbb
"""


def graph_semantics(g):
  """Convert a networkx.DiGraph to compounds and reactions for grid_land."""
  compounds = {}
  reactions = {}
  for node, attributes in g.nodes.items():
    if attributes.get("reaction"):
      reactants = [e[0] for e in g.in_edges(node)]
      products = [e[1] for e in g.out_edges(node)]
      reactions[node] = create_reaction(reactants, products, attributes)
    if not attributes.get("reaction"):
      compounds[node] = create_compound(attributes)

  return compounds, reactions


def create_reaction(reactants, products, attributes):
  # TODO(b/192926758): support fixedSwapOrder = False, in that case, pass
  # reactants# and products as a dictionary mapping to the number required (not
  # a list with possibly repeated entries like the current version).
  return {
      "reactants": reactants,
      "products": products,
      "fixedSwapOrder": attributes.get("fixedSwapOrder", True),
      "priority": attributes.get("priority", 1),
  }


def create_compound(attributes):
  """Convert node attributes to dictionary structure needed for a compound."""
  data = {
      # Use black color if none provided.
      "color": attributes.get("color", (0, 0, 0)),
      "properties": {
          # Use (0, 0) for structure if none provided,
          "structure": attributes.get("structure", (0, 0)),
      },
  }
  for k, v in attributes.items():
    data[k] = v
  return data


def add_system_nodes(g):
  """Add several nodes that must always be present for the system to function.

  Args:
    g: (nx.DiGraph): directed graph representing the reaction system.
  """
  g.add_nodes_from([
      # Add a node for the "empty" compound.
      ("empty", {"color": EARTH_COLOR,
                 "reactivity": "low"}),
      # Add a node for the "activated" compound.
      ("activated", {"color": WHITE_COLOR,
                     "immovable": True}),
      # Add unused nodes that serve only to make all standard groups valid so
      # their corresponding updater can be created.
      ("_unused_a", {"reactivity": "low"}),
      ("_unused_b", {"reactivity": "medium"}),
      ("_unused_c", {"reactivity": "high"})
  ])


def add_compounds_to_prefabs_dictionary(prefabs,
                                        compounds,
                                        reactivity_levels,
                                        sprites=False,
                                        default_reaction_radius=None,
                                        default_reaction_query_type=None,
                                        priority_mode=False):
  """Add compounds."""
  for compound_name in compounds.keys():
    prefabs[compound_name] = create_cell_prefab(
        compound_name,
        compounds,
        reactivity_levels,
        sprites=sprites,
        default_reaction_radius=default_reaction_radius,
        default_reaction_query_type=default_reaction_query_type,
        priority_mode=priority_mode)
  return prefabs


def multiply_tuple(color_tuple, factor):
  if len(color_tuple) == 3:
    return tuple([int(np.min([x * factor, 255])) for x in color_tuple])
  elif len(color_tuple) == 4:
    return tuple([int(np.min([x * factor])) for x in color_tuple])


def create_cell_prefab(compound_name, compounds, reactivity_levels,
                       sprites=False, default_reaction_radius=None,
                       default_reaction_query_type=None, priority_mode=False):
  """Create prefab for a cell object initially set to state=`compound_name`."""
  state_configs = []
  states_to_properties = {}
  sprite_colors = []
  query_configs = {}
  for compound, attributes in compounds.items():
    groups = []
    if "reactivity" in attributes:
      reactivity_group = attributes["reactivity"]
      groups.append(reactivity_group)
    if "immovable" in attributes and attributes["immovable"]:
      groups.append("immovables")
    if "query_config" in attributes:
      query_configs[compound] = attributes["query_config"]

    state_config = {
        "state": compound,
        "sprite": compound,
        "layer": "lowerPhysical",
        "groups": groups  + ["spawnPoints"],
    }
    state_configs.append(state_config)
    states_to_properties[compound] = attributes["properties"]
    sprite_colors.append(attributes["color"])

  # Configure the Reactant component.
  reactivities = {}
  for key, value in reactivity_levels.items():
    reactivities[key] = value

  if sprites:
    def get_palette(sprite_color):
      if len(sprite_color) == 3:
        x_color = EARTH_COLOR[0:3]
        a_color = (252, 252, 252)
      elif len(sprite_color) == 4:
        x_color = EARTH_COLOR
        a_color = (252, 252, 252, 255)
      return {
          "x": x_color,
          "a": a_color,
          "b": sprite_color,
          "c": multiply_tuple(sprite_color, 0.2),
          "d": sprite_color
      }
    appearance_kwargs = {
        "renderMode": "ascii_shape",
        "spriteNames": list(compounds.keys()),
        "spriteShapes": [DIAMOND_SHAPE] * len(sprite_colors),
        "palettes": [get_palette(color) for color in sprite_colors],
        "noRotates": [True] * len(sprite_colors),
    }
    # Must ensure "empty" and "activated" are not given the diamond sprite.
    for i, compound in enumerate(appearance_kwargs["spriteNames"]):
      if compound in ["empty", "activated"]:
        appearance_kwargs["spriteShapes"][i] = SQUARE_SHAPE
  else:
    appearance_kwargs = {
        "spriteNames": list(compounds.keys()),
        "spriteRGBColors": sprite_colors,
    }

  prefab = {
      "name": "cell",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": compound_name,
                  "stateConfigs": state_configs,
              }
          },
          {
              "component": "Transform",
              "kwargs": {
                  "position": (0, 0),
                  "orientation": "N"
              }
          },
          {
              "component": "Appearance",
              "kwargs": appearance_kwargs
          },
          {
              "component": "Cell",
              "kwargs": {
                  "numCellStates": len(state_configs),
                  "statesToProperties": states_to_properties,
                  # The radius over which to search for neighbors on every step.
                  "radius": default_reaction_radius,
                  # Query according to L1 (diamond) or L2 (disc) norm.
                  "queryType": default_reaction_query_type,
                  # Layers on which to search for neighbors on every step.
                  "interactionLayers": ["lowerPhysical", "overlay"],
                  # You can override query properties on a per state basis.
                  "stateSpecificQueryConfig": query_configs,
              },
          },
          {
              "component": "Reactant",
              "kwargs": {
                  "name": "Reactant",
                  "reactivities": reactivities,
                  "priorityMode": priority_mode,
              }
          },
          {
              "component": "Product",
              "kwargs": {
                  "name": "Product",
              }
          },
      ]
  }
  return prefab


def create_stomach(compounds, reactivity_levels, default_reaction_radius=None,
                   default_reaction_query_type=None, priority_mode=False):
  """Construct prefab for an avatar's stomach object."""
  stomach_prefix = "stomach_"
  state_configs = []
  states_to_properties = {}
  sprite_colors = []
  query_configs = {}
  for compound, attributes in compounds.items():
    groups = []
    if "reactivity" in attributes:
      reactivity_group = (stomach_prefix +
                          attributes["reactivity"])
      groups.append(reactivity_group)
    if "immovable" in attributes and attributes["immovable"]:
      groups.append("immovables")
    if "query_config" in attributes:
      query_configs[compound] = attributes["query_config"]
    state_config = {
        "state": compound,
        "sprite": compound + "_stomach",
        "layer": "overlay",
        "groups": groups,
    }
    state_configs.append(state_config)
    states_to_properties[compound] = attributes["properties"]
    sprite_colors.append(attributes["color"])

  # Configure the Reactant component.
  reactivities = {}
  for key, value in reactivity_levels.items():
    reactivities[stomach_prefix + key] = value

  prefab = {
      "name": "avatar_stomach",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "preInit",
                  "stateConfigs": state_configs +
                                  [{"state": "preInit"}],
              }
          },
          {
              "component": "Transform",
              "kwargs": {
                  "position": (0, 0),
                  "orientation": "N"
              }
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [key + "_stomach" for key in compounds.keys()],
                  "spriteShapes": [STOMACH] * len(sprite_colors),
                  # color 0 is a dark navy blue. This is overwritten below.
                  "palettes": [shapes.get_palette(sprite_colors[i])
                               for i in range(len(sprite_colors))],
                  "noRotates": [False] * len(sprite_colors)
              },
          },
          {
              "component": "AvatarStomach",
              "kwargs": {
                  "playerIndex": -1,  # player index to be overwritten.
                  "preInitState": "preInit",
                  "initialState": "empty",
                  "waitState": "stomachWait"
              }
          },
          {
              "component": "Cell",
              "kwargs": {
                  "numCellStates": len(state_configs),
                  "statesToProperties": states_to_properties,
                  # The radius over which to search for neighbors on every step.
                  "radius": default_reaction_radius,
                  # Query according to L1 (diamond) or L2 (disc) norm.
                  "queryType": default_reaction_query_type,
                  # Layers on which to search for neighbors on every step.
                  "interactionLayers": ["lowerPhysical", "overlay"],
                  # You can override query properties on a per state basis.
                  "stateSpecificQueryConfig": query_configs,
              },
          },
          {
              "component": "Reactant",
              "kwargs": {
                  "name": "Reactant",
                  "reactivities": reactivities,
                  "priorityMode": priority_mode,
              }
          },
          {
              "component": "Product",
              "kwargs": {
                  "name": "Product",
              }
          },
      ]
  }
  return prefab


def create_avatar(rewarding_reactions):
  """Create an avatar prefab rewarded by reactions in `rewarding_reactions`."""
  prefab = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "player",
                  "stateConfigs": [
                      {"state": "player",
                       "layer": "upperPhysical",
                       "sprite": "Avatar",
                       "contact": "avatar",
                       "groups": ["players"]},

                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
              "kwargs": {
                  "position": (0, 0),
                  "orientation": "N"
              }
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["Avatar"],
                  "spriteShapes": [shapes.AVATAR_DEFAULT],
                  # color 0 is a dark navy blue. This is overwritten below.
                  "palettes": [shapes.get_palette(colors.palette[0])],
                  "noRotates": [False]
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": -1,  # player index to be overwritten.
                  "spawnGroup": "spawnPoints",
                  "aliveState": "player",
                  "waitState": "playerWait",
                  "actionOrder": ["move", "turn", "ioAction"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": 4},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "ioAction": {"default": 0, "min": 0, "max": 1},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  }
              }
          },
          {
              "component": "IOBeam",
              "kwargs": {
                  "cooldownTime": 2,
              }
          },
          {
              "component": "ReactionsToRewards",
              "kwargs": {
                  # Specify rewards for specific reactions.
                  "rewardingReactions": rewarding_reactions
              }
          },
          {
              "component": "LocationObserver",
              "kwargs": {
                  "objectIsAvatar": True,
                  "alsoReportOrientation": True
              }
          },
      ]
  }
  return prefab


def create_avatar_constant_self_view(
    rewarding_reactions, player_idx: int,
    target_sprite_self: Dict[str, Any]) -> Dict[str, Any]:
  """Create an avatar prefab rewarded by reactions in `rewarding_reactions`."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1

  # Setup the self vs other sprite mapping.
  source_sprite_self = "Avatar" + str(lua_index)
  custom_sprite_map = {source_sprite_self: target_sprite_self["name"]}

  live_state_name = "player{}".format(lua_index)
  avatar_object = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": live_state_name,
                  "stateConfigs": [
                      {"state": live_state_name,
                       "layer": "upperPhysical",
                       "sprite": source_sprite_self,
                       "contact": "avatar",
                       "groups": ["players"]},

                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
              "kwargs": {
                  "position": (0, 0),
                  "orientation": "N"
              }
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [source_sprite_self],
                  "spriteShapes": [shapes.AVATAR_DEFAULT],
                  "palettes": [shapes.get_palette(colors.palette[player_idx])],
                  "noRotates": [False]
              }
          },
          {
              "component": "AdditionalSprites",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "customSpriteNames": [target_sprite_self["name"]],
                  "customSpriteShapes": [target_sprite_self["shape"]],
                  "customPalettes": [target_sprite_self["palette"]],
                  "customNoRotates": [target_sprite_self["noRotate"]],
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "spawnGroup": "spawnPoints",
                  "aliveState": live_state_name,
                  "waitState": "playerWait",
                  "actionOrder": ["move", "turn", "ioAction"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": 4},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "ioAction": {"default": 0, "min": 0, "max": 1},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  },
                  "spriteMap": custom_sprite_map,
              }
          },
          {
              "component": "IOBeam",
              "kwargs": {
                  "cooldownTime": 2,
              }
          },
          {
              "component": "ReactionsToRewards",
              "kwargs": {
                  # Specify rewards for specific reactions.
                  "rewardingReactions": rewarding_reactions
              }
          },
          {
              "component": "LocationObserver",
              "kwargs": {
                  "objectIsAvatar": True,
                  "alsoReportOrientation": True
              }
          },
      ]
  }
  return avatar_object


def create_scene(reactions, stochastic_episode_ending=False):
  """Construct the global scene prefab."""
  scene = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {
              "component": "Transform",
              "kwargs": {
                  "position": (0, 0),
                  "orientation": "N"
              },
          },
          {
              "component": "ReactionAlgebra",
              "kwargs": {
                  "reactions": reactions
              }
          },
          {
              "component": "GlobalMetricTracker",
              "kwargs": {
                  "name": "GlobalMetricTracker",
              }
          },
      ]
  }
  if stochastic_episode_ending:
    scene["components"].append({
        "component": "StochasticIntervalEpisodeEnding",
        "kwargs": {
            "minimumFramesPerEpisode": 1000,
            "intervalLength": 100,  # Set equal to unroll length.
            "probabilityTerminationPerInterval": 0.2
        }
    })
  return scene



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tools to help with parsing and procedurally generating ascii maps."""

from collections.abc import Mapping, Sequence
from typing import Any, Union


def a_or_b_with_odds(a_descriptor: Union[str, Mapping[str, Any]],
                     b_descriptor: Union[str, Mapping[str, Any]],
                     odds: Sequence[int]) -> Mapping[str, Any]:
  """Return a versus b with specified odds.

  Args:
      a_descriptor: One possibility. May be either a string or a dict that can
          be read by the map parser.
      b_descriptor: The other possibility. May be either a string or a dict that
          can be read by the map parser.
      odds: odds[0] is the number of outcomes where a is returned. odds[1] is
          the number of outcomes where b is returned. Thus the probability of
          returning a is odds[0] / sum(odds) and the probability of returning
          b is odds[1] / sum(odds).

  Returns:
      The dict descriptor that can be used with the map parser to sample either
      a or b at the specified odds.
  """
  a_odds, b_odds = odds
  choices = [a_descriptor] * a_odds + [b_descriptor] * b_odds
  return {"type": "choice", "list": choices}



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A simple human player for testing `externality_mushrooms`.

Use `WASD` keys to move the character around.
Use `Q and E` to turn the character.
Use `SPACE` to fire the zapper.
Use `TAB` to switch between players.
"""

import argparse
import json

from ml_collections import config_dict

from meltingpot.configs.substrates import externality_mushrooms__dense
from meltingpot.human_players import level_playing_utils


environment_configs = {
    'externality_mushrooms__dense': externality_mushrooms__dense,
}

_ACTION_MAP = {
    'move': level_playing_utils.get_direction_pressed,
    'turn': level_playing_utils.get_turn_pressed,
    'fireZap': level_playing_utils.get_space_key_pressed,
}


def verbose_fn(env_timestep, player_index, current_player_index):
  """Print using this function once enabling the option --verbose=True."""
  lua_index = player_index + 1
  ate_hihe = env_timestep.observation[f'{lua_index}.ATE_MUSHROOM_HIHE']
  ate_fize = env_timestep.observation[f'{lua_index}.ATE_MUSHROOM_FIZE']
  ate_zife = env_timestep.observation[f'{lua_index}.ATE_MUSHROOM_ZIFE']
  destroyed_hihe = env_timestep.observation[
      f'{lua_index}.DESTROYED_MUSHROOM_HIHE']
  destroyed_fize = env_timestep.observation[
      f'{lua_index}.DESTROYED_MUSHROOM_FIZE']
  destroyed_zife = env_timestep.observation[
      f'{lua_index}.DESTROYED_MUSHROOM_ZIFE']
  at_least_one_nonzero = (ate_hihe + ate_fize + ate_zife +
                          destroyed_hihe + destroyed_fize + destroyed_zife)
  # Only print observations from player 0.
  if player_index == current_player_index and at_least_one_nonzero > 0:
    print(
        f'player: {player_index} --- \n' +
        f'  ate_hihe: {ate_hihe} \n' +
        f'  ate_fize: {ate_fize} \n' +
        f'  ate_zife: {ate_zife} \n' +
        f'  destroyed_hihe: {destroyed_hihe} \n' +
        f'  destroyed_fize: {destroyed_fize} \n' +
        f'  destroyed_zife: {destroyed_zife} \n'
    )


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument(
      '--level_name', type=str, default='externality_mushrooms__dense',
      choices=environment_configs.keys(),
      help='Level name to load')
  parser.add_argument(
      '--observation', type=str, default='RGB', help='Observation to render')
  parser.add_argument(
      '--settings', type=json.loads, default={}, help='Settings as JSON string')
  # Activate verbose mode with --verbose=True.
  parser.add_argument(
      '--verbose', type=bool, default=False, help='Print debug information')
  # Activate events printing mode with --print_events=True.
  parser.add_argument(
      '--print_events', type=bool, default=False, help='Print events')

  args = parser.parse_args()
  env_module = environment_configs[args.level_name]
  env_config = env_module.get_config()
  with config_dict.ConfigDict(env_config).unlocked() as env_config:
    roles = env_config.default_player_roles
    env_config.lab2d_settings = env_module.build(roles, env_config)
  level_playing_utils.run_episode(
      args.observation, args.settings, _ACTION_MAP,
      env_config, level_playing_utils.RenderType.PYGAME,
      verbose_fn=verbose_fn if args.verbose else None,
      print_events=args.print_events)


if __name__ == '__main__':
  main()



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Scenario factory."""

import concurrent
import random
import threading
from typing import Callable, Collection, List, Mapping, Sequence

import chex
import dm_env
import reactivex
from reactivex import subject

from meltingpot.utils.policies import policy as policy_lib


def _step_fn(policy: policy_lib.Policy,
             lock: threading.Lock) -> Callable[[dm_env.TimeStep], int]:
  """Threadsafe stateful step function where the state is encapsulated.

  Args:
    policy: the underlying policy to use.
    lock: a lock that controls access to the policy.

  Returns:
    A step function that returns an action in response to a timestep.
  """
  with lock:
    state = policy.initial_state()

  def step(timestep: dm_env.TimeStep) -> int:
    nonlocal state
    with lock:
      action, state = policy.step(timestep=timestep, prev_state=state)
    return action

  return step


@chex.dataclass(frozen=True)  # works with tree.
class PopulationObservables:
  """Observables for a population.

  Attributes:
    names: emits the names of the sampled population on a reset.
    action: emits actions sent to the substrate by the poulation.
    timestep: emits timesteps sent from the substrate to the population.
  """
  names: reactivex.Observable[Sequence[str]]
  action: reactivex.Observable[Sequence[int]]
  timestep: reactivex.Observable[dm_env.TimeStep]


class Population:
  """A population of policies to use in a scenario."""

  def __init__(
      self,
      *,
      policies: Mapping[str, policy_lib.Policy],
      names_by_role: Mapping[str, Collection[str]],
      roles: Sequence[str]) -> None:
    """Initializes the population.

    Args:
      policies: the policies to sample from (with replacement) each episode.
        Will be closed when the Population is closed.
      names_by_role: dict mapping role to bot names that can fill it.
      roles: specifies which role should fill the corresponding player slot.
    """
    self._policies = dict(policies)
    self._names_by_role = {
        role: tuple(set(names)) for role, names in names_by_role.items()}
    self._roles = tuple(roles)

    self._locks = {name: threading.Lock() for name in self._policies}
    self._executor = concurrent.futures.ThreadPoolExecutor(
        max_workers=len(roles))
    self._step_fns: List[Callable[[dm_env.TimeStep], int]] = []
    self._action_futures: List[concurrent.futures.Future[int]] = []

    self._names_subject = subject.Subject()
    self._action_subject = subject.Subject()
    self._timestep_subject = subject.Subject()
    self._observables = PopulationObservables(  # pylint: disable=unexpected-keyword-arg
        names=self._names_subject,
        action=self._action_subject,
        timestep=self._timestep_subject,
    )

  def close(self):
    """Closes the population."""
    for future in self._action_futures:
      future.cancel()
    self._executor.shutdown(wait=False)
    for policy in self._policies.values():
      policy.close()
    self._names_subject.on_completed()
    self._action_subject.on_completed()
    self._timestep_subject.on_completed()

  def _sample_names(self) -> Sequence[str]:
    """Returns a sample of policy names for the population."""
    return [random.choice(self._names_by_role[role]) for role in self._roles]

  def reset(self) -> None:
    """Resamples the population."""
    names = self._sample_names()
    self._names_subject.on_next(names)
    self._step_fns = [
        _step_fn(policy=self._policies[name], lock=self._locks[name])
        for name in names
    ]
    for future in self._action_futures:
      future.cancel()
    self._action_futures.clear()

  def send_timestep(self, timestep: dm_env.TimeStep) -> None:
    """Sends timestep to population for asynchronous processing.

    Args:
      timestep: The substrate timestep for the population.

    Raises:
      RuntimeError: previous action has not been awaited.
    """
    if self._action_futures:
      raise RuntimeError('Previous action not retrieved.')
    self._timestep_subject.on_next(timestep)
    for n, step_fn in enumerate(self._step_fns):
      bot_timestep = timestep._replace(
          observation=timestep.observation[n], reward=timestep.reward[n])
      future = self._executor.submit(step_fn, bot_timestep)
      self._action_futures.append(future)

  def await_action(self) -> Sequence[int]:
    """Waits for the population action in response to last timestep.

    Returns:
      The action for the population.

    Raises:
      RuntimeError: no timestep has been sent.
    """
    if not self._action_futures:
      raise RuntimeError('No timestep sent.')
    actions = tuple(future.result() for future in self._action_futures)
    self._action_futures.clear()
    self._action_subject.on_next(actions)
    return actions

  def observables(self) -> PopulationObservables:
    """Returns the observables for the population."""
    return self._observables



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Puppeteer test utilities."""

from typing import Any, Iterator, Iterable, Mapping, Optional, Sequence, TypeVar

import dm_env

from meltingpot.utils.puppeteers import puppeteer as puppeteer_lib

GOAL_KEY = puppeteer_lib._GOAL_OBSERVATION_KEY  # pylint: disable=protected-access
State = TypeVar('State')


def step_many(
    puppeteer: puppeteer_lib.Puppeteer[State],
    timesteps: Iterable[dm_env.TimeStep],
    state: Optional[State] = None,
) -> Iterator[tuple[dm_env.TimeStep, State]]:
  """Yields multiple puppeteeer steps."""
  if state is None:
    state = puppeteer.initial_state()
  for timestep in timesteps:
    transformed_timestep, state = puppeteer.step(timestep, state)
    yield transformed_timestep, state


def goals_from_timesteps(
    puppeteer: puppeteer_lib.Puppeteer[State],
    timesteps: Iterable[dm_env.TimeStep],
    state: Optional[State] = None,
) -> tuple[Sequence[puppeteer_lib.PuppetGoal], State]:
  """Returns puppet goals for each timestep."""
  goals = []
  for timestep, state in step_many(puppeteer, timesteps, state):
    goals.append(timestep.observation[GOAL_KEY])
  return goals, state


def episode_timesteps(
    observations: Sequence[Mapping[str, Any]]) -> Iterator[dm_env.TimeStep]:
  """Yields an episode timestep for each observation."""
  for n, observation in enumerate(observations):
    if n == 0:
      yield dm_env.restart(observation=observation)
    elif n == len(observations) - 1:
      yield dm_env.termination(observation=observation, reward=0)
    else:
      yield dm_env.transition(observation=observation, reward=0)


def goals_from_observations(
    puppeteer: puppeteer_lib.Puppeteer[State],
    observations: Sequence[Mapping[str, Any]],
    state: Optional[State] = None,
) -> tuple[Sequence[puppeteer_lib.PuppetGoal], State]:
  """Returns puppet goals from an episode of the provided observations."""
  timesteps = episode_timesteps(observations)
  return goals_from_timesteps(puppeteer, timesteps, state)



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test utilities for substrates."""

from absl.testing import parameterized


class SubstrateTestCase(parameterized.TestCase):
  """Base class for tests of substrates."""

  def assert_step_matches_specs(self, env):
    """Asserts that env accepts an action permitted by its spec.

    Args:
      env: environment to check.

    Raises:
      AssertionError: the env doesn't match its spec.
    """
    env.reset()

    action = [spec.maximum for spec in env.action_spec()]
    try:
      timestep = env.step(action)
    except Exception:  # pylint: disable=broad-except
      self.fail(f'Failure when passing action {action!r}.')

    try:
      env.discount_spec().validate(timestep.discount)
    except ValueError:
      self.fail('Discount does not match spec.')

    reward_spec = env.reward_spec()
    if len(reward_spec) != len(timestep.reward):
      self.fail(f'Spec is length {len(reward_spec)} but reward is length '
                f'{len(timestep.reward)}.')
    for n, spec in enumerate(reward_spec):
      try:
        spec.validate(timestep.reward[n])
      except ValueError:
        self.fail(f'Reward {n} does not match spec.')

    observations = timestep.observation
    observation_specs = env.observation_spec()
    if len(observation_specs) != len(observations):
      self.fail(f'Spec is length {len(observation_specs)} but observations '
                f'are length {len(observations)}')
    for n, (observation, spec) in enumerate(
        zip(observations, observation_specs)):
      if set(spec) != set(observation):
        self.fail(f'Observation {n} keys {set(observation)!r} do not match '
                  f'spec keys {set(observation)!r}.')
      for key in spec:
        try:
          spec[key].validate(observation[key])
        except ValueError:
          self.fail(f'Observation {n} key {key!r} does not match spec.')



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrapper that rebuilds the Lab2d environment on every reset."""

from typing import Callable

import dm_env
import dmlab2d

from meltingpot.utils.substrates.wrappers import base


class ResetWrapper(base.Lab2dWrapper):
  """Wrapper that rebuilds the environment on reset."""

  def __init__(self, build_environment: Callable[[], dmlab2d.Environment]):
    """Initializes the object.

    Args:
      build_environment: Called to build the underlying environment.
    """
    env = build_environment()
    super().__init__(env)
    self._rebuild_environment = build_environment
    self._reset = False

  def reset(self) -> dm_env.TimeStep:
    """Rebuilds the environment and calls reset on it."""
    if self._reset:
      self._env.close()
      self._env = self._rebuild_environment()
    else:
      # Don't rebuild on very first reset call (it's inefficient).
      self._reset = True
    return super().reset()



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for wrappers.

Wrappers are assumed to own the wrapped environment and that they have the
**only** reference to it. This means that they will:

1.   Close the environment when they close.
2.   Modify the environment specs and timesteps inplace.
"""

import dmlab2d


class Lab2dWrapper(dmlab2d.Environment):
  """Base class for wrappers of dmlab2d.Environments."""

  def __init__(self, env):
    """Initializes the wrapper.

    Args:
      env: An environment to wrap. This environment will be closed with this
        wrapper.
    """
    self._env = env

  def reset(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.reset(*args, **kwargs)

  def step(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.step(*args, **kwargs)

  def reward_spec(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.reward_spec(*args, **kwargs)

  def discount_spec(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.discount_spec(*args, **kwargs)

  def observation_spec(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.observation_spec(*args, **kwargs)

  def action_spec(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.action_spec(*args, **kwargs)

  def close(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.close(*args, **kwargs)

  def observation(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.observation(*args, **kwargs)

  def events(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.events(*args, **kwargs)

  def list_property(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.list_property(*args, **kwargs)

  def write_property(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.write_property(*args, **kwargs)

  def read_property(self, *args, **kwargs) -> ...:
    """See base class."""
    return self._env.read_property(*args, **kwargs)



# Copyright 2023 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile

from absl.testing import absltest
import cv2
import dm_env
import numpy as np

from meltingpot.utils.evaluation import evaluation


def _as_timesteps(frames):
  first, *mids, last = frames
  yield dm_env.restart(observation=[{'WORLD.RGB': first}])
  for frame in mids:
    yield dm_env.transition(observation=[{'WORLD.RGB': frame}], reward=0)
  yield dm_env.termination(observation=[{'WORLD.RGB': last}], reward=0)


def _get_frames(path):
  capture = cv2.VideoCapture(path)
  while capture.isOpened():
    ret, bgr_frame = capture.read()
    if not ret:
      break
    rgb_frame = cv2.cvtColor(bgr_frame, cv2.COLOR_BGR2RGB)
    yield rgb_frame
  capture.release()


FRAME_SHAPE = (4, 8)
ZERO = np.zeros(FRAME_SHAPE, np.uint8)
EYE = np.eye(*FRAME_SHAPE, dtype=np.uint8) * 255
RED_EYE = np.stack([EYE, ZERO, ZERO], axis=-1)
GREEN_EYE = np.stack([ZERO, EYE, ZERO], axis=-1)
BLUE_EYE = np.stack([ZERO, ZERO, EYE], axis=-1)


class EvaluationTest(absltest.TestCase):

  def test_video_subject(self):
    video_path = None
    step_written = None

    def save_path(path):
      nonlocal video_path
      video_path = path

    tempdir = tempfile.mkdtemp()
    assert os.path.exists(tempdir)
    # Use lossless compression for test.
    subject = evaluation.VideoSubject(tempdir, extension='avi', codec='png ')
    subject.subscribe(on_next=save_path)

    frames = [RED_EYE, GREEN_EYE, BLUE_EYE]
    for n, timestep in enumerate(_as_timesteps(frames)):
      subject.on_next(timestep)
      if step_written is None and video_path is not None:
        step_written = n

    with self.subTest('video_exists'):
      self.assertTrue(video_path and os.path.exists(video_path))

    with self.subTest('written_on_final_step'):
      self.assertEqual(step_written, 2)

    with self.subTest('contents'):
      written = list(_get_frames(video_path))
      np.testing.assert_equal(written, frames)

  def test_return_subject(self):
    episode_return = None
    step_written = None

    def save_return(ret):
      nonlocal episode_return
      episode_return = ret

    subject = evaluation.ReturnSubject()
    subject.subscribe(on_next=save_return)

    timesteps = [
        dm_env.restart(observation=[{}])._replace(reward=[0, 0]),
        dm_env.transition(observation=[{}], reward=[2, 4]),
        dm_env.termination(observation=[{}], reward=[1, 3]),
    ]
    for n, timestep in enumerate(timesteps):
      subject.on_next(timestep)
      if step_written is None and episode_return is not None:
        step_written = n

    with self.subTest('written_on_final_step'):
      self.assertEqual(step_written, 2)

    with self.subTest('contents'):
      np.testing.assert_equal(episode_return, [3, 7])

if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Puppeteers for *_in_the_matrix."""

from collections.abc import Mapping, Sequence
import dataclasses
import random
from typing import Optional, TypeVar

import dm_env
import numpy as np
import tree

from meltingpot.utils.puppeteers import puppeteer

State = TypeVar("State")
Observation = Mapping[str, tree.Structure[np.ndarray]]


def get_inventory(timestep: dm_env.TimeStep) -> np.ndarray:
  """Returns player's current inventory."""
  return timestep.observation["INVENTORY"]


def get_partner_interaction_inventory(
    timestep: dm_env.TimeStep) -> Optional[np.ndarray]:
  """Returns the partner inventory from previous interaction."""
  _, partner_inventory = timestep.observation["INTERACTION_INVENTORIES"]
  if np.all(partner_inventory < 0):
    return None  # No interaction occurred.
  else:
    return partner_inventory


def has_interaction(timestep: dm_env.TimeStep) -> bool:
  """Returns True if the timestep contains an interaction."""
  return get_partner_interaction_inventory(timestep) is not None


def max_resource_and_margin(inventory: np.ndarray) -> tuple[int, int]:
  """Returns the index of the maximum resource and the margin of its lead."""
  sorted_resources = np.argsort(inventory)
  maximum_resource = sorted_resources[-1]
  margin = (
      int(inventory[sorted_resources[-1]]) -
      int(inventory[sorted_resources[-2]]))
  return maximum_resource, margin


def has_collected_sufficient(
    inventory: np.ndarray,
    resource: int,
    margin: int,
) -> bool:
  """Returns True if a sufficient amount of the resource has been collected.

  Args:
    inventory: the inventory of collected resources.
    resource: the index of the resource being collected.
    margin: the required margin for "sufficiency".
  """
  max_resource, current_margin = max_resource_and_margin(inventory)
  return max_resource == resource and current_margin >= margin


def partner_max_resource(timestep: dm_env.TimeStep) -> Optional[int]:
  """Returns partner's maximum resource at previous interaction."""
  partner_inventory = get_partner_interaction_inventory(timestep)
  if partner_inventory is None:
    return None  # No interaction occurred.
  resource, margin = max_resource_and_margin(partner_inventory)
  if margin == 0:
    return None  # Intent is unclear (no unique maximum).
  else:
    return resource


def tremble(tremble_probability: float):
  """Returns True if the hand trembles."""
  return random.random() < tremble_probability


@dataclasses.dataclass(frozen=True)
class Resource:
  """A resource that can be collected by a puppet.

  Attributes:
    index: the index of the resource in the INVENTORY vector.
    collect_goal: the goal that directs the puppet to collect the resource.
    interact_goal: the goal that directs the puppet to interact with another
      player while playing the resource.
  """
  index: int
  collect_goal: puppeteer.PuppetGoal
  interact_goal: puppeteer.PuppetGoal

  def __eq__(self, obj):
    if not isinstance(obj, Resource):
      return NotImplemented
    else:
      return self is obj

  def __hash__(self):
    return hash(id(self))


def collect_or_interact_puppet_timestep(
    timestep: dm_env.TimeStep,
    target: Resource,
    margin: int,
) -> dm_env.TimeStep:
  """Returns a timestep for a *_in_the_matrix puppet.

  Args:
    timestep: the timestep without any goal added.
    target: the resource for the collector to target.
    margin: the threshold at which the puppet switches from collecting to
      interacting.

  Returns:
    A timestep with a goal added for the puppet. If the puppet has already
    collected enough of the targeted resource, will add the resource's
    interact_goal. Otherwise will add the resource's collect_goal.
  """
  inventory = get_inventory(timestep)
  if has_collected_sufficient(inventory, target.index, margin):
    goal = target.interact_goal
  else:
    goal = target.collect_goal
  return puppeteer.puppet_timestep(timestep, goal)


class Specialist(puppeteer.Puppeteer[tuple[()]]):
  """Puppeteer that targets a single resource."""

  def __init__(self, *, target: Resource, margin: int) -> None:
    """Initializes the puppeteer.

    Args:
      target: the resource to target.
      margin: the margin at which the specialist will switch from collecting to
        interacting.
    """
    self._target = target
    if margin > 0:
      self._margin = margin
    else:
      raise ValueError("Margin must be positive.")

  def initial_state(self) -> tuple[()]:
    """See base class."""
    return ()

  def step(self, timestep: dm_env.TimeStep,
           prev_state: tuple[()]) -> tuple[dm_env.TimeStep, tuple[()]]:
    """See base class."""
    timestep = collect_or_interact_puppet_timestep(
        timestep, self._target, self._margin)
    return timestep, prev_state


class AlternatingSpecialist(puppeteer.Puppeteer[int]):
  """Puppeteer that cycles targeted resource on a fixed schedule."""

  def __init__(self,
               *,
               targets: Sequence[Resource],
               interactions_per_target: int,
               margin: int) -> None:
    """Initializes the puppeteer.

    Args:
      targets: circular sequence of resources to target. Targets correspond to
        pure strategies in the underlying matrix game.
      interactions_per_target: how many interactions to select each target
        before switching to the next one in the `targets` sequence.
      margin: Try to collect `margin` more of the target resource than the other
        resources before interacting.
    """
    if targets:
      self._targets = tuple(targets)
    else:
      raise ValueError("targets must not be empty")

    if interactions_per_target > 0:
      self._interactions_per_target = interactions_per_target
    else:
      raise ValueError("interactions_per_target must be positive.")

    if margin > 0:
      self._margin = margin
    else:
      raise ValueError("margin must be positive.")

  def initial_state(self) -> int:
    """See base class."""
    return 0

  def step(self, timestep: dm_env.TimeStep,
           prev_state: int) -> tuple[dm_env.TimeStep, int]:
    """See base class."""
    if timestep.first():
      prev_state = self.initial_state()

    if has_interaction(timestep):
      total_interactions = prev_state + 1
    else:
      total_interactions = prev_state

    target_index = (total_interactions // self._interactions_per_target) % len(
        self._targets)
    target = self._targets[target_index]

    timestep = collect_or_interact_puppet_timestep(
        timestep, target, self._margin)

    return timestep, total_interactions


class ScheduledFlip(puppeteer.Puppeteer[int]):
  """Puppeteer that targets one resource then switches to another."""

  def __init__(
      self,
      *,
      threshold: int,
      initial_target: Resource,
      final_target: Resource,
      initial_margin: int,
      final_margin: int,
  ) -> None:
    """Initializes the puppeteer.

    Args:
      threshold: Switch targeted resource once this many interactions have
        occurred.
      initial_target: The initial resource to target.
      final_target: The resource to target after the switch.
      initial_margin: How much more of the target resource to collect before
        interacting.
      final_margin: The margin after the flip.
    """
    self._initial_target = initial_target
    self._final_target = final_target

    if threshold > 0:
      self._threshold = threshold
    else:
      raise ValueError("threshold must be positive.")

    if initial_margin > 0:
      self._initial_margin = initial_margin
    else:
      raise ValueError("initial_margin must be positive.")

    if final_margin > 0:
      self._final_margin = final_margin
    else:
      raise ValueError("final_margin must be positive.")

  def initial_state(self) -> int:
    """See base class."""
    return 0

  def step(self, timestep: dm_env.TimeStep,
           prev_state: int) -> tuple[dm_env.TimeStep, int]:
    """See base class."""
    if timestep.first():
      prev_state = self.initial_state()

    if has_interaction(timestep):
      total_interactions = prev_state + 1
    else:
      total_interactions = prev_state

    if total_interactions < self._threshold:
      timestep = collect_or_interact_puppet_timestep(
          timestep, self._initial_target, self._initial_margin)
    else:
      timestep = collect_or_interact_puppet_timestep(
          timestep, self._final_target, self._final_margin)

    return timestep, total_interactions


class GrimTrigger(puppeteer.Puppeteer[int]):
  """Puppeteer for a grim trigger.

  This bot will always try to play cooperate until other players have defected
  against it more than `threshold` times. After enduring `threshold` defections,
  it switches to a triggered mode where it always plays defect. It never leaves
  this mode, i.e. it is grim. It defects in all future interactions, not only
  those interactions with the players who originally defected on it.
  """

  def __init__(
      self,
      *,
      threshold: int,
      cooperate_resource: Resource,
      defect_resource: Resource,
      margin: int,
  ) -> None:
    """Initializes the puppeteer.

    Args:
      threshold: How many defections cause this agent to switch to its triggered
        mode. Once triggered it will try to defect in all future interactions.
      cooperate_resource: the cooperation resource.
      defect_resource: the defection resource.
      margin: Try to collect `margin` more of the target resource than the other
        resource before interacting.
    """
    if threshold > 0:
      self._threshold = threshold
    else:
      raise ValueError("threshold must be positive")

    self._cooperate_resource = cooperate_resource
    self._defect_resource = defect_resource

    if margin > 0:
      self._margin = margin
    else:
      raise ValueError("margin must be positive")

  def initial_state(self) -> int:
    """See base class."""
    return 0

  def step(self, timestep: dm_env.TimeStep,
           prev_state: int) -> tuple[dm_env.TimeStep, int]:
    """See base class."""
    if timestep.first():
      prev_state = self.initial_state()

    partner_resource = partner_max_resource(timestep)
    partner_defected = partner_resource == self._defect_resource.index
    if partner_defected:
      partner_defections = prev_state + 1
    else:
      partner_defections = prev_state

    if partner_defections < self._threshold:
      timestep = collect_or_interact_puppet_timestep(
          timestep, self._cooperate_resource, self._margin)
    else:
      timestep = collect_or_interact_puppet_timestep(
          timestep, self._defect_resource, self._margin)
    return timestep, partner_defections


class TitForTat(puppeteer.Puppeteer[bool]):
  """Puppeteer for a tit-for-tat bot.

  This bot will always try to cooperate if its partner cooperated in the last
  round and defect if its partner defected in the last round. It cooperates
  on the first round.

  Important note: this puppeteer function assumes there is only one other player
  in the game. So it only makes sense for two player substrates like those we
  called *_in_the_matrix__repeated.
  """

  def __init__(
      self,
      *,
      cooperate_resource: Resource,
      defect_resource: Resource,
      margin: int,
      tremble_probability: float,
  ) -> None:
    """Initializes the puppeteer.

    Args:
      cooperate_resource: the cooperation resource.
      defect_resource: the defection resource.
      margin: Try to collect `margin` more of the target resource than the other
        resource before interacting.
      tremble_probability: When deciding to cooperate/defect, switch to
        defect/cooperate with this probability.
    """
    self._cooperate_resource = cooperate_resource
    self._defect_resource = defect_resource

    if margin > 0:
      self._margin = margin
    else:
      raise ValueError("margin must be positive")

    if 0 <= tremble_probability <= 1:
      self._tremble_probability = tremble_probability
    else:
      raise ValueError("tremble_probability must be a probability.")

  def initial_state(self) -> bool:
    """See base class."""
    is_cooperative = True if not tremble(self._tremble_probability) else False
    return is_cooperative

  def step(self, timestep: dm_env.TimeStep,
           prev_state: bool) -> tuple[dm_env.TimeStep, bool]:
    """See base class."""
    if timestep.first():
      prev_state = self.initial_state()

    partner_resource = partner_max_resource(timestep)
    partner_defected = partner_resource == self._defect_resource.index
    partner_cooperated = partner_resource == self._cooperate_resource.index

    if partner_cooperated:
      is_cooperative = True if not tremble(self._tremble_probability) else False
    elif partner_defected:
      is_cooperative = False if not tremble(self._tremble_probability) else True
    else:
      is_cooperative = prev_state

    if is_cooperative:
      timestep = collect_or_interact_puppet_timestep(
          timestep, self._cooperate_resource, self._margin)
    else:
      timestep = collect_or_interact_puppet_timestep(
          timestep, self._defect_resource, self._margin)
    return timestep, is_cooperative


@dataclasses.dataclass(frozen=True)
class CorrigableState:
  """State of Corrigable puppeteer.

  Attributes:
    partner_defections: the number of times the partner has defected.
    is_cooperative: whether the puppeteer is currently cooperating (as opposed
       to defecting).
  """
  partner_defections: int
  is_cooperative: bool


class Corrigible(puppeteer.Puppeteer[CorrigableState]):
  """Puppeteer that defects until you punish it, then switches to tit-for-tat.

  Important note: this puppeteer function assumes there is only one other player
  in the game. So it only makes sense for two player substrates like those we
  called *_in_the_matrix__repeated.
  """

  def __init__(
      self,
      threshold: int,
      cooperate_resource: Resource,
      defect_resource: Resource,
      margin: int,
      tremble_probability: float,
  ) -> None:
    """Initializes the puppeteer.

    Args:
      threshold: How many times this bot must be punished for it to change its
        behavior from 'always defect' to 'tit-for-tat'.
      cooperate_resource: the cooperation resource.
      defect_resource: the defection resource.
      margin: Try to collect `margin` more of the target resource than the other
        resource before interacting.
      tremble_probability: Once playing tit-for-tat, when deciding to
        cooperate/defect, switch to defect/cooperate with this probability.
    """
    if threshold > 0:
      self._threshold = threshold
    else:
      raise ValueError("threshold must be positive.")

    self._cooperate_resource = cooperate_resource
    self._defect_resource = defect_resource

    if margin > 0:
      self._margin = margin
    else:
      raise ValueError("margin must be positive")

    if 0 <= tremble_probability <= 1:
      self._tremble_probability = tremble_probability
    else:
      raise ValueError("tremble_probability must be a probability.")

  def initial_state(self) -> CorrigableState:
    """See base class."""
    return CorrigableState(partner_defections=0, is_cooperative=False)

  def step(
      self,
      timestep: dm_env.TimeStep,
      prev_state: CorrigableState,
  ) -> tuple[dm_env.TimeStep, CorrigableState]:
    """See base class."""
    if timestep.first():
      prev_state = self.initial_state()

    partner_resource = partner_max_resource(timestep)
    partner_defected = partner_resource == self._defect_resource.index
    partner_cooperated = partner_resource == self._cooperate_resource.index

    if partner_defected:
      partner_defections = prev_state.partner_defections + 1
      switching_now = partner_defections == self._threshold
    else:
      partner_defections = prev_state.partner_defections
      switching_now = False

    insufficiently_punished = partner_defections < self._threshold
    if insufficiently_punished:
      is_cooperative = False
    elif switching_now or partner_cooperated:
      is_cooperative = True if not tremble(self._tremble_probability) else False
    elif partner_defected:
      is_cooperative = False if not tremble(self._tremble_probability) else True
    else:
      is_cooperative = prev_state.is_cooperative

    if is_cooperative:
      timestep = collect_or_interact_puppet_timestep(
          timestep, self._cooperate_resource, self._margin)
    else:
      timestep = collect_or_interact_puppet_timestep(
          timestep, self._defect_resource, self._margin)
    next_state = CorrigableState(
        is_cooperative=is_cooperative, partner_defections=partner_defections)
    return timestep, next_state


class RespondToPrevious(puppeteer.Puppeteer[Resource]):
  """Puppeteer for responding to opponents previous move.

  At the start of an episode, RespondToPrevious targets a random resource up
  until the first interaction occurs. Thereafter RespondToPrevious selects the
  resource to target based on the maximum resource held by the coplayer at the
  last interaction. If the coplayer held no single maximum resource,
  RespondToPrevious will continue to target the resource it was previously
  targeting.
  """

  def __init__(
      self,
      responses: Mapping[Resource, Resource],
      margin: int,
  ) -> None:
    """Initializes the puppeteer.

    Args:
      responses: Mapping from the maximum resource in the partner inventory to
        the resource to target in response.
      margin: Try to collect `margin` more of the target resource than the other
        resource before interacting.
    """
    self._responses = {
        resource.index: response for resource, response in responses.items()
    }
    if margin > 0:
      self._margin = margin
    else:
      raise ValueError("margin must be positive.")

  def initial_state(self) -> Resource:
    """See base class."""
    return random.choice(list(self._responses.values()))

  def step(
      self,
      timestep: dm_env.TimeStep,
      prev_state: Resource,
  ) -> tuple[dm_env.TimeStep, Resource]:
    """See base class."""
    if timestep.first():
      prev_state = self.initial_state()
    partner_resource = partner_max_resource(timestep)
    response = self._responses.get(partner_resource, prev_state)
    timestep = collect_or_interact_puppet_timestep(
        timestep, response, self._margin)
    return timestep, response



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrapper that converts the DMLab2D specs into lists of action/observation."""

from collections.abc import Collection, Iterator, Mapping, Sequence
from typing import TypeVar

import dm_env
import numpy as np

from meltingpot.utils.substrates.wrappers import observables

T = TypeVar("T")


def _player_observations(observations: Mapping[str, T], suffix: str,
                         num_players: int) -> Iterator[T]:
  """Yields observations for each player.

  Args:
    observations: dmlab2d observations source to check.
    suffix: suffix of player key to return.
    num_players: the number of players.
  """
  for player_index in range(num_players):
    try:
      value = observations[f"{player_index + 1}.{suffix}"]
    except KeyError:
      pass
    else:
      if isinstance(value, dm_env.specs.Array):
        value = value.replace(name=suffix)
      yield player_index, value


class Wrapper(observables.ObservableLab2dWrapper):
  """Wrapper that converts the environment to multiplayer lists.

  Ensures:
  -   observations are returned as lists of dictionary observations
  -   rewards are returned as lists of scalars
  -   actions are received as lists of dictionary observations
  -   discounts are never None
  """

  def __init__(self, env,
               individual_observation_names: Collection[str],
               global_observation_names: Collection[str]):
    """Constructor.

    Args:
      env: environment to wrap. When this wrapper closes env will also be
        closed.
      individual_observation_names: the per-player observations to make
        available to the players.
      global_observation_names: the observations that are available to all
        players and analytics.
    """
    super().__init__(env)
    self._num_players = self._get_num_players()
    self._individual_observation_suffixes = set(individual_observation_names)
    self._global_observation_names = set(global_observation_names)

  def _get_num_players(self) -> int:
    """Returns maximum player index in dmlab2d action spec."""
    action_spec_keys = super().action_spec().keys()
    lua_player_indices = (int(key.split(".", 1)[0]) for key in action_spec_keys)
    return max(lua_player_indices)

  def _get_observations(
      self, source: Mapping[str, T]) -> Sequence[Mapping[str, T]]:
    """Returns multiplayer observations from dmlab2d observations.

    Args:
      source: dmlab2d observations source to check.
    """
    player_observations = [{} for i in range(self._num_players)]
    for suffix in self._individual_observation_suffixes:
      for i, value in _player_observations(source, suffix, self._num_players):
        player_observations[i][suffix] = value
    for name in self._global_observation_names:
      value = source[name]
      for i in range(self._num_players):
        player_observations[i][name] = value
    return player_observations

  def _get_rewards(self, source: Mapping[str, T]) -> Sequence[T]:
    """Returns multiplayer rewards from dmlab2d observations.

    Args:
      source: dmlab2d observations source to check.
    """
    rewards = [None] * self._num_players
    for i, value in _player_observations(source, "REWARD", self._num_players):
      rewards[i] = value
    return rewards

  def _get_timestep(self, source: dm_env.TimeStep) -> dm_env.TimeStep:
    """Returns multiplayer timestep from dmlab2d observations.

    Args:
      source: dmlab2d observations source to check.
    """
    return dm_env.TimeStep(
        step_type=source.step_type,
        reward=self._get_rewards(source.observation),
        discount=0. if source.discount is None else source.discount,
        observation=self._get_observations(source.observation))

  def _get_action(self, source: Sequence[Mapping[str, T]]) -> Mapping[str, T]:
    """Returns dmlab2 action from multiplayer actions.

    Args:
      source: multiplayer actions.
    """
    dmlab2d_actions = {}
    for player_index, action in enumerate(source):
      for key, value in action.items():
        dmlab2d_actions[f"{player_index + 1}.{key}"] = value
    return dmlab2d_actions

  def reset(self) -> dm_env.TimeStep:
    """See base class."""
    timestep = super().reset()
    return self._get_timestep(timestep)

  def step(
      self, actions: Sequence[Mapping[str, np.ndarray]]) -> dm_env.TimeStep:
    """See base class."""
    action = self._get_action(actions)
    timestep = super().step(action)
    return self._get_timestep(timestep)

  def observation(self) -> Sequence[Mapping[str, np.ndarray]]:
    """See base class."""
    observation = super().observation()
    return self._get_observations(observation)

  def action_spec(self) -> Sequence[Mapping[str, dm_env.specs.Array]]:
    """See base class."""
    source = super().action_spec()
    action_spec = [{} for _ in range(self._num_players)]
    for key, spec in source.items():
      lua_player_index, suffix = key.split(".", 1)
      player_index = int(lua_player_index) - 1
      action_spec[player_index][suffix] = spec.replace(name=suffix)
    return action_spec

  def observation_spec(self) -> Sequence[Mapping[str, dm_env.specs.Array]]:
    """See base class."""
    source = super().observation_spec()
    return self._get_observations(source)

  def reward_spec(self) -> Sequence[dm_env.specs.Array]:
    """See base class."""
    source = super().observation_spec()
    return self._get_rewards(source)



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for shapes."""

from absl.testing import absltest
from absl.testing import parameterized

from meltingpot.utils.substrates import shapes


class ShapesTest(parameterized.TestCase):

  @parameterized.parameters([
      ["""
a
b
c
""", """
c
b
a
"""], ["""
""", """
"""], ["""
abc
def
ghi
""", """
ghi
def
abc
"""],
  ])
  def test_flip_vertical(self, original, expected):
    actual = shapes.flip_vertical(original)
    self.assertEqual(actual, expected)

  @parameterized.parameters([
      ["""
a
b
c
""", """
a
b
c
"""], ["""
""", """
"""], ["""
abc
def
ghi
""", """
cba
fed
ihg
"""],
  ])
  def test_flip_horizontal(self, original, expected):
    actual = shapes.flip_horizontal(original)
    self.assertEqual(actual, expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for hidden_agenda.

Example video: https://youtu.be/voJWckiOh5k

A social deduction (sometimes also called 'hidden role') game where players can
have one of two roles: crewmates and impostors. All players with the same role
are in the same team. The game is a zero-sum competitive game across teams, that
is, all players in the team get the same reward, and their reward is the
negative of the reward for the other team. The roles are hidden (hence the name)
and must be inferred from observations. The crewmates have a numeric advantage
(4 players), while the impostor (1 player) have an information advantage (they
know the roles of every player).

Players can move around a 2D world which contains gems that can be picked up by
walking over them, a deposit in the center of the map (a grate) where collected
gems can be dropped, and a voting room where rounds of deliberation occur (see
below).

Crewmates can carry up to two gems in their inventory at the same time. The gems
must be deposited in the grate before more gems can be collected. Impostors have
a freezing beam that they can use to freeze crewmates. Frozen crewmates are
unable to move or take any action for the rest of the episode.

After a predefined time (200 steps) or whenever an impostor fires its beam
within the field of view of another player (and is not frozen by it) a
deliberation phase starts. During deliberation, players are teleported to the
deliberation room, where they cannot take any movement or firing actions. Their
actions are limited to voting actions. Votes can be for any of the player
indices (0 to 4), 'abstain', or 'no-vote' (if the player is frozen or voted
out). The deliberation phase lasts 25 steps and players are able to change their
vote at any point. If there is a simple majority in the last step of the
deliberation, the player is voted out and removed from the game. Players can
observe the voting of every other player as a special observation called
`VOTING`. Episodes last up to 3000 steps.

The game has several win conditions:

1.  The crewmembers deposit enough gems (32). Crewmates win.
2.  The impostor is voted out during the deliberation phase. Crewmates win.
3.  There is only one crewmate active (i.e. not voted out nor frozen). Impostor
    wins.

If neither of the above conditions are met before the episode ends, the game is
considered a tie, and players get zero reward.
"""
import copy
from typing import Any, Mapping, Sequence

from ml_collections import config_dict
from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

# This substrate only makes sense with exactly five players.
MANDATED_NUM_PLAYERS = 5

# Hidden Agenda palette for player slots
HIDDEN_AGENDA_COLORS = [
    (37, 133, 190),  # light blue
    (133, 37, 190),  # indigo
    (255, 95, 10),  # orange
    (37, 190, 133),  # sea green
    (220, 40, 110),  # salmon pink
    (180, 180, 0),  # golden yellow
    (133, 190, 37),  # lime green
    (135, 73, 124),  # dark pink-purple
    (140, 115, 105),  # light brown
]

# Dictionary mapping the characters in the map with prefabs.
CHAR_PREFAB_MAP = {
    "*": {"type": "all", "list": ["checkered_flooring", "spawn_point"]},
    "V": {"type": "all", "list": ["tiled_flooring1", "voting_spawn_point"]},
    "D": {"type": "all", "list": ["tiled_floor", "teleport_spawn_point"]},
    "F": "nw_wall_corner",
    "7": "ne_wall_corner",
    "J": "se_wall_corner",
    "L": "sw_wall_corner",
    "[": "w_ship_solid_wall",
    "]": "e_ship_solid_wall",
    "^": "n_ship_solid_wall",
    "v": "s_ship_solid_wall",
    "-": "wall_north",
    "T": "tcoupling_n",
    "Z": "tcoupling_e",
    "i": "tcoupling_s",
    "t": "tcoupling_w",
    "|": "wall_west",
    "f": "fill",
    ",": {"type": "all", "list": ["nw_grate", "gem_deposit"]},
    "_": {"type": "all", "list": ["n_grate", "gem_deposit"]},
    ";": {"type": "all", "list": ["ne_grate", "gem_deposit"]},
    "!": {"type": "all", "list": ["w_grate", "gem_deposit"]},
    "=": {"type": "all", "list": ["inner_grate", "gem_deposit"]},
    "1": {"type": "all", "list": ["e_grate", "gem_deposit"]},
    "+": {"type": "all", "list": ["se_grate", "gem_deposit"]},
    "'": {"type": "all", "list": ["s_grate", "gem_deposit"]},
    "`": {"type": "all", "list": ["sw_grate", "gem_deposit"]},
    "/": {"type": "all", "list": ["tiled_floor", "glass_wall"]},
    "n": "tiled_floor",
    "U": "tiled_flooring1",
    "u": "tiled_flooring2",
    "m": "metal_flooring",
    "e": "metal_panel_flooring",
    "x": "checkered_flooring",
    "w": "wood_flooring",
    "~": "threshold",
    "%": {"type": "all", "list": ["metal_panel_flooring", "gem"]},
    "@": {"type": "all", "list": ["metal_flooring", "gem"]},
    "&": {"type": "all", "list": ["wood_flooring", "gem"]},
    "#": {"type": "all", "list": ["tiled_floor", "gem"]},
}

ASCII_MAP = """
F----------^^-------^^----------7
|@mmmmmmmmm[]DDDDDDD[]mmmmmmmmmm|
|mmmmmm@mmm[]///////[]mm@mmm@mmm|
|m@mmmm@mmm|UuVuVuVuU|mmmmm@mmm@|
|mmmm@mm@mm|uVuUuUuVu|mmmmm@mmmm|
|m@mmm@mmmm|UuVuUuVuU|mm@mmmmm@m|
|mm@m@mm@mm|uUuVuVuUu|mm@mm@mmmm|
t-~~~~~~~~-i---------i-~~~~~~~~-Z
|xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx|
|xxxxxxxxxx*xx,___;xx*xxxxxxxxxx|
|xxxxxxxxxx**x!===1x**xxxxxxxxxx|
|xxxxxxxxxx**x!===1x**xxxxxxxxxx|
|xxxxxxxxxx*xx`'''+xx*xxxxxxxxxx|
|xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx|
t-~~~~~~~~-^^^^^^^^^^^-~~~~~~~~-Z
|mmmm@mm@mm[fffffffff]mm@mmmm@mm|
|mmmmmm@mmm[fffffffff]mm@m@mmmmm|
|m@mmmmmm@m[fffffffff]@mmmmm@mmm|
|mmmmm@mmmm[fffffffff]mm@mmmmmm@|
|m@mmmm@mm@[fffffffff]mm@mmmm@mm|
|mmm@mm@mmm[fffffffff]@mmmmmmmmm|
L----------vvvvvvvvvvv----------J
"""


COMPASS = ["N", "E", "S", "W"]

# Aesthetic components for Hidden Agenda

NW_WALL_CORNER = {
    "name": "nw_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "nw_wall_corner",
                "stateConfigs": [{
                    "state": "nw_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "NwWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NwWallCorner"],
                "spriteShapes": [shapes.NW_SHIP_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

NE_WALL_CORNER = {
    "name": "ne_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "ne_wall_corner",
                "stateConfigs": [{
                    "state": "ne_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "NeWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NeWallCorner"],
                "spriteShapes": [shapes.NE_SHIP_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SE_WALL_CORNER = {
    "name": "se_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "se_wall_corner",
                "stateConfigs": [{
                    "state": "se_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "SeWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["SeWallCorner"],
                "spriteShapes": [shapes.SE_SHIP_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SW_WALL_CORNER = {
    "name": "sw_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "sw_wall_corner",
                "stateConfigs": [{
                    "state": "sw_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "SwWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["SwWallCorner"],
                "spriteShapes": [shapes.SW_SHIP_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

WALL_NORTH = {
    "name": "wall_north",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall_north",
                "stateConfigs": [{
                    "state": "wall_north",
                    "layer": "upperPhysical",
                    "sprite": "WallNorth",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallNorth"],
                "spriteShapes": [shapes.NS_SHIP_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

TCOUPLING_E = {
    "name": "tcoupling_e",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "tcoupling_e",
                "stateConfigs": [{
                    "state": "tcoupling_e",
                    "layer": "upperPhysical",
                    "sprite": "TcouplingE",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["TcouplingE"],
                "spriteShapes": [shapes.SHIP_WALL_TCOUPLING_E],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

TCOUPLING_W = {
    "name": "tcoupling_w",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "tcoupling_w",
                "stateConfigs": [{
                    "state": "tcoupling_w",
                    "layer": "upperPhysical",
                    "sprite": "TcouplingW",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["TcouplingW"],
                "spriteShapes": [shapes.SHIP_WALL_TCOUPLING_W],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

TCOUPLING_N = {
    "name": "tcoupling_n",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "tcoupling_n",
                "stateConfigs": [{
                    "state": "tcoupling_n",
                    "layer": "upperPhysical",
                    "sprite": "TcouplingN",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["TcouplingN"],
                "spriteShapes": [shapes.SHIP_WALL_TCOUPLING_N],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

TCOUPLING_S = {
    "name": "tcoupling_s",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "tcoupling_s",
                "stateConfigs": [{
                    "state": "tcoupling_s",
                    "layer": "upperPhysical",
                    "sprite": "TcouplingS",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["TcouplingS"],
                "spriteShapes": [shapes.SHIP_WALL_TCOUPLING_S],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

W_SHIP_SOLID_WALL = {
    "name": "w_ship_solid_wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "w_ship_solid_wall",
                "stateConfigs": [{
                    "state": "w_ship_solid_wall",
                    "layer": "upperPhysical",
                    "sprite": "WShipSolidWall",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WShipSolidWall"],
                "spriteShapes": [shapes.W_SHIP_SOLID_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

N_SHIP_SOLID_WALL = {
    "name": "n_ship_solid_wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "n_ship_solid_wall",
                "stateConfigs": [{
                    "state": "n_ship_solid_wall",
                    "layer": "upperPhysical",
                    "sprite": "NShipSolidWall",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NShipSolidWall"],
                "spriteShapes": [shapes.N_SHIP_SOLID_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

E_SHIP_SOLID_WALL = {
    "name": "e_ship_solid_wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "e_ship_solid_wall",
                "stateConfigs": [{
                    "state": "e_ship_solid_wall",
                    "layer": "upperPhysical",
                    "sprite": "EShipSolidWall",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["EShipSolidWall"],
                "spriteShapes": [shapes.E_SHIP_SOLID_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

S_SHIP_SOLID_WALL = {
    "name": "s_ship_solid_wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "s_ship_solid_wall",
                "stateConfigs": [{
                    "state": "s_ship_solid_wall",
                    "layer": "upperPhysical",
                    "sprite": "SShipSolidWall",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["SShipSolidWall"],
                "spriteShapes": [shapes.S_SHIP_SOLID_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

WALL_WEST = {
    "name": "wall_west",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall_west",
                "stateConfigs": [{
                    "state": "wall_west",
                    "layer": "upperPhysical",
                    "sprite": "WallWest",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallWest"],
                "spriteShapes": [shapes.EW_SHIP_WALL],
                "palettes": [shapes.SHIP_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

NW_GRATE = {
    "name": "nw_grate",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "nw_grate",
                "stateConfigs": [{
                    "state": "nw_grate",
                    "layer": "background",
                    "sprite": "nw_grate",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["nw_grate"],
                "spriteShapes": [shapes.NW_GRATE],
                "palettes": [shapes.GRATE_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

N_GRATE = {
    "name": "n_grate",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "n_grate",
                "stateConfigs": [{
                    "state": "n_grate",
                    "layer": "background",
                    "sprite": "n_grate",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["n_grate"],
                "spriteShapes": [shapes.N_GRATE],
                "palettes": [shapes.GRATE_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

NE_GRATE = {
    "name": "ne_grate",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "ne_grate",
                "stateConfigs": [{
                    "state": "ne_grate",
                    "layer": "background",
                    "sprite": "ne_grate",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["ne_grate"],
                "spriteShapes": [shapes.NE_GRATE],
                "palettes": [shapes.GRATE_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

W_GRATE = {
    "name": "w_grate",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "w_grate",
                "stateConfigs": [{
                    "state": "w_grate",
                    "layer": "background",
                    "sprite": "w_grate",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["w_grate"],
                "spriteShapes": [shapes.W_GRATE],
                "palettes": [shapes.GRATE_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

INNER_GRATE = {
    "name": "inner_grate",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "inner_grate",
                "stateConfigs": [{
                    "state": "inner_grate",
                    "layer": "background",
                    "sprite": "inner_grate",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["inner_grate"],
                "spriteShapes": [shapes.INNER_GRATE],
                "palettes": [shapes.GRATE_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

E_GRATE = {
    "name": "e_grate",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "e_grate",
                "stateConfigs": [{
                    "state": "e_grate",
                    "layer": "background",
                    "sprite": "e_grate",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["e_grate"],
                "spriteShapes": [shapes.E_GRATE],
                "palettes": [shapes.GRATE_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SE_GRATE = {
    "name": "se_grate",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "se_grate",
                "stateConfigs": [{
                    "state": "se_grate",
                    "layer": "background",
                    "sprite": "se_grate",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["se_grate"],
                "spriteShapes": [shapes.SE_GRATE],
                "palettes": [shapes.GRATE_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

S_GRATE = {
    "name": "s_grate",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "s_grate",
                "stateConfigs": [{
                    "state": "s_grate",
                    "layer": "background",
                    "sprite": "s_grate",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["s_grate"],
                "spriteShapes": [shapes.S_GRATE],
                "palettes": [shapes.GRATE_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SW_GRATE = {
    "name": "sw_grate",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "sw_grate",
                "stateConfigs": [{
                    "state": "sw_grate",
                    "layer": "background",
                    "sprite": "sw_grate",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["sw_grate"],
                "spriteShapes": [shapes.SW_GRATE],
                "palettes": [shapes.GRATE_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

GLASS_WALL = {
    "name": "glass_wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "glass_wall",
                "stateConfigs": [{
                    "state": "glass_wall",
                    "layer": "upperPhysical",
                    "sprite": "glass_wall",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["glass_wall"],
                "spriteShapes": [shapes.GLASS_WALL],
                "palettes": [shapes.GLASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

FILL = {
    "name": "fill",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "fill",
                "stateConfigs": [{
                    "state": "fill",
                    "layer": "upperPhysical",
                    "sprite": "fill",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["fill"],
                "spriteShapes": [shapes.FILL],
                "palettes": [{"i": (58, 68, 102, 255),}],
                "noRotates": [False]
            }
        },
    ]
}

TILED_FLOOR = {
    "name": "tiled_floor",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "tiled_floor",
                "stateConfigs": [{
                    "state": "tiled_floor",
                    "layer": "background",
                    "sprite": "tiled_floor",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["tiled_floor"],
                "spriteShapes": [shapes.TILED_FLOOR_GREY],
                "palettes": [{"o": (204, 199, 192, 255),
                              "-": (194, 189, 182, 255),}],
                "noRotates": [False]
            }
        },
    ]
}

WOOD_FLOOR = {
    "name": "wood_floor",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wood_floor",
                "stateConfigs": [{
                    "state": "wood_floor",
                    "layer": "background",
                    "sprite": "wood_floor",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["wood_floor"],
                "spriteShapes": [shapes.WOOD_FLOOR],
                "palettes": [shapes.WOOD_FLOOR_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

METAL_FLOORING = {
    "name": "metal_flooring",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "metal_flooring",
                "stateConfigs": [{
                    "state": "metal_flooring",
                    "layer": "background",
                    "sprite": "metal_flooring",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["metal_flooring"],
                "spriteShapes": [shapes.METAL_TILE],
                "palettes": [shapes.METAL_FLOOR_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

METAL_PANEL_FLOORING = {
    "name": "metal_panel_flooring",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "metal_panel_flooring",
                "stateConfigs": [{
                    "state": "metal_panel_flooring",
                    "layer": "background",
                    "sprite": "metal_panel_flooring",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["metal_panel_flooring"],
                "spriteShapes": [shapes.METAL_PANEL],
                "palettes": [shapes.METAL_PANEL_FLOOR_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

CHECKERED_FLOORING = {
    "name": "checkered_flooring",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "checkered_flooring",
                "stateConfigs": [{
                    "state": "checkered_flooring",
                    "layer": "background",
                    "sprite": "checkered_flooring",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["checkered_flooring"],
                "spriteShapes": [shapes.CHECKERED_TILE],
                "palettes": [{"X": (120, 108, 108, 255),
                              "x": (115, 103, 103, 255),}],
                "noRotates": [False]
            }
        },
    ]
}

TILED_FLOORING1 = {
    "name": "tiled_flooring1",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "tiled_flooring1",
                "stateConfigs": [{
                    "state": "tiled_flooring1",
                    "layer": "background",
                    "sprite": "tiled_flooring1",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["tiled_flooring1"],
                "spriteShapes": [shapes.TILE1],
                "palettes": [shapes.TILE_FLOOR_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

TILED_FLOORING2 = {
    "name": "tiled_flooring2",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "tiled_flooring2",
                "stateConfigs": [{
                    "state": "tiled_flooring2",
                    "layer": "background",
                    "sprite": "tiled_flooring2",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["tiled_flooring2"],
                "spriteShapes": [shapes.TILE2],
                "palettes": [shapes.TILE_FLOOR_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

THRESHOLD = {
    "name": "threshold",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "threshold",
                "stateConfigs": [{
                    "state": "threshold",
                    "layer": "background",
                    "sprite": "threshold",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["threshold"],
                "spriteShapes": [shapes.THRESHOLD],
                "palettes": [{"X": (92, 95, 92, 255),
                              "x": (106, 108, 106, 255),}],
                "noRotates": [False]
            }
        },
    ]
}

# Functional components for Hidden Agenda

SPAWN_POINT = {
    "name": "spawn_point",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "spawnPoint",
                "stateConfigs": [{
                    "state": "spawnPoint",
                    "groups": ["spawnPoints"],
                }],
            }
        },
        {
            "component": "Transform",
        },
    ]
}

VOTING_SPAWN_POINT = {
    "name": "voting_spawn_point",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "votingSpawnPoint",
                "stateConfigs": [{
                    "state": "votingSpawnPoint",
                    "groups": ["votingSpawnPoints"],
                }],
            }
        },
        {
            "component": "Transform",
        },
    ]
}

TELEPORT_SPAWN_POINT = {
    "name": "teleport_spawn_point",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "teleportSpawnPoint",
                "stateConfigs": [{
                    "state": "teleportSpawnPoint",
                    "groups": ["teleportSpawnPoints"],
                }],
            }
        },
        {
            "component": "Transform",
        },
    ]
}


def get_gem_prefab(crewmate_pseudoreward: float):
  return {
      "name": "gem",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "gem",
                  "stateConfigs": [{
                      "state": "gem",
                      "layer": "lowerPhysical",
                      "sprite": "Gem",
                  }, {
                      "state": "gemWait",
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["Gem",],
                  "spriteShapes": [shapes.SMALL_SPHERE],
                  "palettes": [shapes.MOULD_PALETTE],
                  "noRotates": [True],
              }
          },
          {
              "component": "Collectable",
              "kwargs": {
                  "liveState": "gem",
                  "waitState": "gemWait",
                  "rewardForCollecting_crewmate": crewmate_pseudoreward,
                  "rewardForCollecting_impostor": 0.0,
                  "regrowRate": 0.001,
              }
          },
      ]
  }


def get_gem_deposit_prefab(crewmate_pseudoreward: float):
  return {
      "name": "gem_deposit",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "gemDeposit",
                  "stateConfigs": [{
                      "state": "gemDeposit",
                      "layer": "lowerPhysical",
                      "sprite": "GemDeposit",
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["GemDeposit"],
                  "spriteShapes": ["x"],
                  "palettes": [shapes.GRATE_PALETTE],
                  "noRotates": [False],
              }
          },
          {
              "component": "Deposit",
              "kwargs": {
                  "crewmateReward": crewmate_pseudoreward,
                  "impostorReward": 0.0,
              }
          }
      ]
  }


def create_player(player_idx: int, role: str, num_players: int,
                  pseudoreward_for_freezing: float,
                  pseudoreward_for_being_frozen: float):
  """Create a prefab for a Player (Impostor or Crewmate).

  Args:
    player_idx: The index of this player.
    role: Whether this player will be a `crewmate` or an `impostor`.
    num_players: The number of players in the environment.
    pseudoreward_for_freezing: Peudoreward given when the impostor successfully
      freezes a crewmate.
    pseudoreward_for_being_frozen: Pseudoreward (usually negative) given to a
      crewmate that was frozen by the impostor.

  Returns:
    A prefab (dictionary) for a Player.
  """
  # Lua is 1-indexed.
  lua_index = player_idx + 1
  live_state_name = f"player{lua_index}"
  avatar_sprite_name = f"avatarSprite{lua_index}"

  if role == "impostor":
    sprite_map = {avatar_sprite_name: f"Player_impostor{lua_index}"}
  else:
    sprite_map = {}

  player = {
      "name": "player",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": live_state_name,
                  "stateConfigs": [
                      {"state": live_state_name,
                       "layer": "upperPhysical",
                       "sprite": avatar_sprite_name,
                       "contact": "avatar"},
                      {"state": "playerWait"},
                      {"state": "playerBody",
                       "layer": "upperPhysical",
                       "sprite": "Player_tagged"},
                  ]
              },
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [avatar_sprite_name, "Player_tagged"],
                  "spriteShapes": [shapes.CUTE_AVATAR,
                                   shapes.CUTE_AVATAR_FROZEN],
                  "palettes": [
                      shapes.get_palette(HIDDEN_AGENDA_COLORS[player_idx]),
                      shapes.get_palette(HIDDEN_AGENDA_COLORS[player_idx])
                  ],
                  "noRotates": [True]
              }
          },
          {
              "component": "AdditionalPlayerSprites",
              "kwargs": {
                  "renderMode":
                      "ascii_shape",
                  "customSpriteNames": [
                      "Player_impostor" + str(i + 1) for i in range(num_players)
                  ],
                  "customSpriteShapes": [shapes.CUTE_AVATAR_W_BUBBLE] *
                                        num_players,
                  "customPalettes": [
                      shapes.get_palette(HIDDEN_AGENDA_COLORS[i])
                      for i in range(num_players)
                  ],
                  "customNoRotates": [True]
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "spawnGroup": "spawnPoints",
                  "aliveState": live_state_name,
                  "waitState": "playerWait",
                  "actionOrder": ["move", "turn", "tag", "vote"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": len(COMPASS)},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "tag": {"default": 0, "min": 0, "max": 1},
                      "vote": {"default": 0, "min": 0, "max": num_players + 1},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  },
                  "spriteMap": sprite_map,
              },
          },
          {
              "component": "Role",
              "kwargs": {
                  "frozenState": "playerBody",
                  "role": role,  # `crewmate` or `impostor`.
              }
          },
          {
              "component": "Inventory",
              "kwargs": {
                  "max_gems": 1,
              }
          },
          {
              "component": "AdditionalObserver",
              "kwargs": {
                  "num_players": num_players,
              }
          },
          {
              "component": "Tagger",
              "kwargs": {
                  "cooldownTime": 50,
                  "beamLength": 2,
                  "beamRadius": 2,
                  "penaltyForBeingTagged": pseudoreward_for_being_frozen,
                  "rewardForTagging": pseudoreward_for_freezing,
                  "removeHitPlayer": "freeze",
              }
          },
          {
              "component": "ReadyToShootObservation",
              "kwargs": {
                  "zapperComponent": "Tagger",
              },
          },
          {
              "component": "Voting",
              "kwargs": {
                  "spawnGroup": "votingSpawnPoints",
                  "votingActive": False,
                  "votingMethod": "deliberation",
                  "votingValues": {},
              }
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    player["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })

  return player


def create_prefabs(
    crewmate_collect_pseudoreward: float,
    crewmate_deposit_pseudoreward: float,
):
  """Create prefabs dictionary from individual prefabs.

  Args:
    crewmate_collect_pseudoreward: Pseudoreward given to crewmates when they
      collect a gem.
    crewmate_deposit_pseudoreward: Pseudoreward given to crewmates when they
      deposit a gem.

  Returns:
    A dictionary of prefabs for components in the environment.
  """

  # PREFABS is a dictionary mapping names to template game objects that can
  # be cloned and placed in multiple locations accoring to an ascii map.
  prefabs = {
      "nw_wall_corner": NW_WALL_CORNER,
      "ne_wall_corner": NE_WALL_CORNER,
      "sw_wall_corner": SW_WALL_CORNER,
      "n_ship_solid_wall": N_SHIP_SOLID_WALL,
      "e_ship_solid_wall": E_SHIP_SOLID_WALL,
      "s_ship_solid_wall": S_SHIP_SOLID_WALL,
      "w_ship_solid_wall": W_SHIP_SOLID_WALL,
      "wall_north": WALL_NORTH,
      "se_wall_corner": SE_WALL_CORNER,
      "tcoupling_e": TCOUPLING_E,
      "tcoupling_w": TCOUPLING_W,
      "tcoupling_n": TCOUPLING_N,
      "tcoupling_s": TCOUPLING_S,
      "wall_west": WALL_WEST,
      "nw_grate": NW_GRATE,
      "n_grate": N_GRATE,
      "ne_grate": NE_GRATE,
      "w_grate": W_GRATE,
      "inner_grate": INNER_GRATE,
      "e_grate": E_GRATE,
      "se_grate": SE_GRATE,
      "s_grate": S_GRATE,
      "sw_grate": SW_GRATE,
      "glass_wall": GLASS_WALL,
      "fill": FILL,
      "tiled_floor": TILED_FLOOR,
      "tiled_flooring1": TILED_FLOORING1,
      "tiled_flooring2": TILED_FLOORING2,
      "wood_flooring": WOOD_FLOOR,
      "metal_flooring": METAL_FLOORING,
      "metal_panel_flooring": METAL_PANEL_FLOORING,
      "checkered_flooring": CHECKERED_FLOORING,
      "threshold": THRESHOLD,
      "gem": get_gem_prefab(crewmate_collect_pseudoreward),
      "spawn_point": SPAWN_POINT,
      "voting_spawn_point": VOTING_SPAWN_POINT,
      "teleport_spawn_point": TELEPORT_SPAWN_POINT,
      "gem_deposit": get_gem_deposit_prefab(crewmate_deposit_pseudoreward),
  }
  return prefabs

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn": 0,  "tag": 0, "vote": 0}
FORWARD    = {"move": 1, "turn": 0,  "tag": 0, "vote": 0}
STEP_RIGHT = {"move": 2, "turn": 0,  "tag": 0, "vote": 0}
BACKWARD   = {"move": 3, "turn": 0,  "tag": 0, "vote": 0}
STEP_LEFT  = {"move": 4, "turn": 0,  "tag": 0, "vote": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "tag": 0, "vote": 0}
TURN_RIGHT = {"move": 0, "turn": 1,  "tag": 0, "vote": 0}
TAG        = {"move": 0, "turn": 0,  "tag": 1, "vote": 0}
# pylint: enable=bad-whitespace
# pyformat: enable


def create_action_set(num_players):
  """Create the action set for the agents."""
  action_set = [
      NOOP,
      FORWARD, BACKWARD,
      STEP_LEFT, STEP_RIGHT,
      TURN_LEFT, TURN_RIGHT,
      TAG,
  ]
  # vote for each player and no-vote.
  for player in range(1, num_players+2):
    vote = copy.deepcopy(NOOP)
    vote["vote"] = player
    action_set.append(vote)
  return action_set


def get_config():
  """Default configuration for the Hidden Agenda level."""
  config = config_dict.ConfigDict()

  # Specify the number of players to particate in each episode (optional).
  config.recommended_num_players = MANDATED_NUM_PLAYERS

  # Configurable pseudorewards.
  # The canonical substrate requires these pseudorewards to be 0. However,
  # you can set them to something else to improve training. For example, you'd
  # use `crewmate_collect_pseudoreward=0.25` and
  # `crewmate_deposit_pseudoreward=0.25` to aid the crewmates in learning to
  # collect and deposit gems. Also, you can set `pseudoreward_for_freezing=1`
  # and `pseudoreward_for_being_frozen=-1` to help players learn about freezing
  # mechanics.
  config.pseudorewards = dict(
      crewmate_collect_pseudoreward=0.0,
      crewmate_deposit_pseudoreward=0.0,
      pseudoreward_for_freezing=0.0,
      pseudoreward_for_being_frozen=0.0,
  )

  config.scene_prefab = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component":
                  "Progress",
              "kwargs":
                  config_dict.ConfigDict({
                      "num_players":
                          MANDATED_NUM_PLAYERS,
                      "goal": 32,
                      "potential_pseudorewards": False,
                      "crewmate_task_reward": 1,
                      "impostor_task_reward": -1,
                      "crewmate_tag_reward": -1,
                      "impostor_tag_reward": 1,
                      "crewmate_vote_reward": 1,
                      "impostor_vote_reward": -1,
                      "incorrect_vote_reward": 0,
                      "correct_vote_reward": 0,
                      "step_reward": 0,
                      "teleport_spawn_group":
                          "teleportSpawnPoints",
                      "voting_params": {
                          "type": "deliberation",
                          "votingPhaseCooldown": 25,
                          "votingFrameFrequency": 200,
                          "taggingTriggerVoting": True,
                      }
                  })
          },
      ]
  }
  # The voting matrix metric is always used.
  metrics = [{
      "name": "VOTING",
      "type": "tensor.DoubleTensor",
      "shape": (MANDATED_NUM_PLAYERS, MANDATED_NUM_PLAYERS + 2),
      "component": "Progress",
      "variable": "votingMatrix",
  }]
  if _ENABLE_DEBUG_OBSERVATIONS:
    config.scene_prefab["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })
    metrics.append({
        "name": "GLOBAL_PROGRESS",
        "type": "tensor.DoubleTensor",
        "shape": (1,),
        "component": "Progress",
        "variable": "progress_bar",
    })
    metrics.append({
        "name": "IDENTITIES",
        "type": "tensor.DoubleTensor",
        "shape": (MANDATED_NUM_PLAYERS,),
        "component": "Progress",
        "variable": "identity_tensor"
    })
    metrics.append({
        "name": "VOTING",
        "type": "tensor.DoubleTensor",
        "shape": (MANDATED_NUM_PLAYERS, MANDATED_NUM_PLAYERS + 2),
        "component": "Progress",
        "variable": "votingMatrix"
    })

  # Add the global metrics reporter
  config.scene_prefab["components"].append({
      "component": "GlobalMetricReporter",
      "kwargs": {
          "metrics": metrics
      }
  })

  # Action set configuration.
  config.action_set = create_action_set(MANDATED_NUM_PLAYERS)

  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "INVENTORY",
      "READY_TO_SHOOT",
      "VOTING",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(config.action_set))
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "INVENTORY": specs.inventory(1),
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      "VOTING": specs.float64(MANDATED_NUM_PLAYERS, MANDATED_NUM_PLAYERS + 2),
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(176, 264),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"crewmate",
                                  "impostor",})
  config.default_player_roles = ("crewmate",) * 4 + ("impostor",)

  return config


def build(
    roles: Sequence[str],
    config: config_dict.ConfigDict,
) -> Mapping[str, Any]:
  """Build the hidden_agenda substrate given player preferences."""
  # Build avatars.
  num_players = len(roles)
  avatar_objects = []
  for player_idx, role in enumerate(roles):
    # Create an avatar with the correct role.
    avatar_objects.append(create_player(
        player_idx=player_idx,
        role=role,
        num_players=num_players,
        pseudoreward_for_freezing=
        config.pseudorewards.pseudoreward_for_freezing,
        pseudoreward_for_being_frozen=
        config.pseudorewards.pseudoreward_for_being_frozen))
  substrate_definition = dict(
      levelName="hidden_agenda",
      levelDirectory="meltingpot/lua/levels",
      maxEpisodeLengthFrames=3000,
      spriteSize=8,
      numPlayers=num_players,
      topology="BOUNDED",  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": ASCII_MAP,
          "gameObjects": avatar_objects,
          "prefabs": create_prefabs(
              config.pseudorewards.crewmate_collect_pseudoreward,
              config.pseudorewards.crewmate_deposit_pseudoreward,
          ),
          "charPrefabMap": CHAR_PREFAB_MAP,
          "scene": config.scene_prefab,
      },
  )
  return substrate_definition



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for human_players."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from ml_collections import config_dict
import pygame

from meltingpot.configs.substrates import allelopathic_harvest__open
from meltingpot.configs.substrates import boat_race__eight_races
from meltingpot.configs.substrates import chemistry__three_metabolic_cycles
from meltingpot.configs.substrates import chemistry__three_metabolic_cycles_with_plentiful_distractors
from meltingpot.configs.substrates import chemistry__two_metabolic_cycles
from meltingpot.configs.substrates import chemistry__two_metabolic_cycles_with_distractors
from meltingpot.configs.substrates import clean_up
from meltingpot.configs.substrates import coins
from meltingpot.configs.substrates import collaborative_cooking__asymmetric
from meltingpot.configs.substrates import commons_harvest__closed
from meltingpot.configs.substrates import coop_mining
from meltingpot.configs.substrates import daycare
from meltingpot.configs.substrates import externality_mushrooms__dense
from meltingpot.configs.substrates import factory_commons__either_or
from meltingpot.configs.substrates import fruit_market__concentric_rivers
from meltingpot.configs.substrates import gift_refinements
from meltingpot.configs.substrates import paintball__capture_the_flag
from meltingpot.configs.substrates import paintball__king_of_the_hill
from meltingpot.configs.substrates import predator_prey__alley_hunt
from meltingpot.configs.substrates import predator_prey__orchard
from meltingpot.configs.substrates import prisoners_dilemma_in_the_matrix__arena
from meltingpot.configs.substrates import territory__rooms
from meltingpot.human_players import level_playing_utils
from meltingpot.human_players import play_allelopathic_harvest
from meltingpot.human_players import play_anything_in_the_matrix
from meltingpot.human_players import play_boat_race
from meltingpot.human_players import play_chemistry
from meltingpot.human_players import play_clean_up
from meltingpot.human_players import play_coins
from meltingpot.human_players import play_collaborative_cooking
from meltingpot.human_players import play_commons_harvest
from meltingpot.human_players import play_coop_mining
from meltingpot.human_players import play_daycare
from meltingpot.human_players import play_externality_mushrooms
from meltingpot.human_players import play_factory_commons
from meltingpot.human_players import play_fruit_market
from meltingpot.human_players import play_gift_refinements
from meltingpot.human_players import play_paintball
from meltingpot.human_players import play_predator_and_prey
from meltingpot.human_players import play_territory


class PlayLevelTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('allelopathic_harvest__open', allelopathic_harvest__open,
       play_allelopathic_harvest),
      ('boat_race__eight_races', boat_race__eight_races, play_boat_race),
      ('chemistry__three_metabolic_cycles', chemistry__three_metabolic_cycles,
       play_chemistry),
      ('chemistry__three_metabolic_cycles_with_plentiful_distractors',
       chemistry__three_metabolic_cycles_with_plentiful_distractors,
       play_chemistry),
      ('chemistry__two_metabolic_cycles', chemistry__two_metabolic_cycles,
       play_chemistry),
      ('chemistry__two_metabolic_cycles_with_distractors',
       chemistry__two_metabolic_cycles_with_distractors, play_chemistry),
      ('clean_up', clean_up, play_clean_up),
      ('coins', coins, play_coins),
      ('collaborative_cooking__asymmetric', collaborative_cooking__asymmetric,
       play_collaborative_cooking),
      ('commons_harvest__closed', commons_harvest__closed,
       play_commons_harvest),
      ('coop_mining', coop_mining, play_coop_mining),
      ('daycare', daycare, play_daycare),
      ('externality_mushrooms__dense', externality_mushrooms__dense,
       play_externality_mushrooms),
      ('factory_commons__either_or', factory_commons__either_or,
       play_factory_commons),
      ('fruit_market__concentric_rivers', fruit_market__concentric_rivers,
       play_fruit_market),
      ('gift_refinements', gift_refinements, play_gift_refinements),
      ('paintball__capture_the_flag', paintball__capture_the_flag,
       play_paintball),
      ('paintball__king_of_the_hill', paintball__king_of_the_hill,
       play_paintball),
      ('predator_prey__alley_hunt', predator_prey__alley_hunt,
       play_predator_and_prey),
      ('predator_prey__orchard', predator_prey__orchard,
       play_predator_and_prey),
      ('prisoners_dilemma_in_the_matrix__arena',
       prisoners_dilemma_in_the_matrix__arena, play_anything_in_the_matrix),
      ('territory__rooms', territory__rooms, play_territory),
  )
  @mock.patch.object(pygame, 'key')
  @mock.patch.object(pygame, 'display')
  @mock.patch.object(pygame, 'event')
  @mock.patch.object(pygame, 'time')
  def test_run_level(
      self, config_module, play_module, unused_k, unused_d, unused_e, unused_t):
    env_module = config_module
    env_config = env_module.get_config()

    with config_dict.ConfigDict(env_config).unlocked() as env_config:
      roles = env_config.default_player_roles
      env_config.lab2d_settings = env_module.build(roles, env_config)

    env_config['lab2d_settings']['maxEpisodeLengthFrames'] = 10
    level_playing_utils.run_episode(
        'RGB', {}, play_module._ACTION_MAP, env_config)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Allelopathic Harvest (open).

Example video: https://youtu.be/Bb0duMG0YF4

This substrate contains three different varieties of berry (red, green, & blue)
and a fixed number of berry patches, which could be replanted to grow any color
variety of berry. The growth rate of each berry variety depends linearly on the
fraction that that color comprises of the total. Players have three planting
actions with which they can replant berries in their chosen color. All players
prefer to eat red berries (reward of 2 per red berry they eat versus a reward
of 1 per other colored berry). Players can achieve higher return by selecting
just one single color of berry to plant, but which one to pick is, in principle,
difficult to coordinate (start-up problem) -- though in this case all prefer
red berries, suggesting a globally rational chioce. They also always prefer to
eat berries over spending time planting (free-rider problem).

Allelopathic Harvest was first described in Koster et al. (2020).

Köster, R., McKee, K.R., Everett, R., Weidinger, L., Isaac, W.S., Hughes, E.,
Duenez-Guzman, E.A., Graepel, T., Botvinick, M. and Leibo, J.Z., 2020.
Model-free conventions in multi-agent reinforcement learning with heterogeneous
preferences. arXiv preprint arXiv:2010.09054.
"""

from meltingpot.configs.substrates import allelopathic_harvest as base_config

OPEN_ASCII_MAP = """
333PPPP12PPP322P32PPP1P13P3P3
1PPPP2PP122PPP3P232121P2PP2P1
P1P3P11PPP13PPP31PPPP23PPPPPP
PPPPP2P2P1P2P3P33P23PP2P2PPPP
P1PPPPPPP2PPP12311PP3321PPPPP
133P2PP2PPP3PPP1PPP2213P112P1
3PPPPPPPPPPPPP31PPPPPP1P3112P
PP2P21P21P33PPPPPPP3PP2PPPP1P
PPPPP1P1P32P3PPP22PP1P2PPPP2P
PPP3PP3122211PPP2113P3PPP1332
PP12132PP1PP1P321PP1PPPPPP1P3
PPP222P12PPPP1PPPP1PPP321P11P
PPP2PPPP3P2P1PPP1P23322PP1P13
23PPP2PPPP2P3PPPP3PP3PPP3PPP2
2PPPP3P3P3PP3PP3P1P3PP11P21P1
21PPP2PP331PP3PPP2PPPPP2PP3PP
P32P2PP2P1PPPPPPP12P2PPP1PPPP
P3PP3P2P21P3PP2PP11PP1323P312
2P1PPPPP1PPP1P2PPP3P32P2P331P
PPPPP1312P3P2PPPP3P32PPPP2P11
P3PPPP221PPP2PPPPPPPP1PPP311P
32P3PPPPPPPPPP31PPPP3PPP13PPP
PPP3PPPPP3PPPPPP232P13PPPPP1P
P1PP1PPP2PP3PPPPP33321PP2P3PP
P13PPPP1P333PPPP2PP213PP2P3PP
1PPPPP3PP2P1PP21P3PPPP231P2PP
1331P2P12P2PPPP2PPP3P23P21PPP
P3P131P3PPP13P1PPP222PPPP11PP
2P3PPPPPPPP2P323PPP2PPP1PPP2P
21PPPPPPP12P23P1PPPPPP13P3P11
"""

build = base_config.build


def get_config():
  """Adjust default configuration."""
  config = base_config.get_config()
  config.ascii_map = OPEN_ASCII_MAP

  config.default_player_roles = (
      ("player_who_likes_red",) * 8 + ("player_who_likes_green",) * 8)

  return config



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for builder.py."""

import copy

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from meltingpot.configs.substrates import running_with_scissors_in_the_matrix__repeated as test_substrate
from meltingpot.utils.substrates import builder


def _get_test_settings():
  config = test_substrate.get_config()
  return test_substrate.build(config, config.default_player_roles)


_TEST_SETTINGS = _get_test_settings()


def _get_lua_randomization_map():
  """Replaces first row of walls with items randomized by Lua."""
  head, line, *tail = _TEST_SETTINGS['simulation']['map'].split('\n')
  # Replace line 1 (walls) with a row of 'a' (items randomized by Lua).
  new_map = '\n'.join([head, 'a' * len(line), *tail])
  return new_map


_LUA_RANDOMIZED_LINE = 1
_LUA_RANDOMIZATION_MAP = _get_lua_randomization_map()


class GeneralTestCase(parameterized.TestCase):

  @parameterized.product(seed=[42, 123, 1337, 12481632])
  def test_seed_causes_determinism(self, seed):
    env1 = self.enter_context(builder.builder(_TEST_SETTINGS, env_seed=seed))
    env2 = self.enter_context(builder.builder(_TEST_SETTINGS, env_seed=seed))
    for episode in range(5):
      obs1 = env1.reset().observation['WORLD.RGB']
      obs2 = env2.reset().observation['WORLD.RGB']
      np.testing.assert_equal(
          obs1, obs2, f'Episode {episode} mismatch: {obs1} != {obs2} ')

  @parameterized.product(seed=[None, 42, 123, 1337, 12481632])
  def test_episodes_are_randomized(self, seed):
    env = self.enter_context(builder.builder(_TEST_SETTINGS, env_seed=seed))

    obs = env.reset().observation['WORLD.RGB']
    for episode in range(4):
      last_obs = obs
      obs = env.reset().observation['WORLD.RGB']
      with self.assertRaises(
          AssertionError,
          msg=f'Episodes {episode} and {episode+1} match: {last_obs} == {obs}'):
        np.testing.assert_equal(last_obs, obs)

  def test_no_seed_causes_nondeterminism(self):
    env1 = self.enter_context(builder.builder(_TEST_SETTINGS, env_seed=None))
    env2 = self.enter_context(builder.builder(_TEST_SETTINGS, env_seed=None))
    for episode in range(5):
      obs1 = env1.reset().observation['WORLD.RGB']
      obs2 = env2.reset().observation['WORLD.RGB']
      with self.assertRaises(
          AssertionError, msg=f'Episode {episode} match: {obs1} == {obs2}'):
        np.testing.assert_equal(obs1, obs2)

  @parameterized.product(seed=[None, 42, 123, 1337, 12481632])
  def test_episodes_are_randomized_in_lua(self, seed):
    lab2d_settings = copy.deepcopy(_TEST_SETTINGS)
    lab2d_settings['simulation']['map'] = _LUA_RANDOMIZATION_MAP
    env = self.enter_context(builder.builder(lab2d_settings, env_seed=seed))

    obs = env.reset().observation['WORLD.RGB'][_LUA_RANDOMIZED_LINE]
    for episode in range(4):
      last_obs = obs
      obs = env.reset().observation['WORLD.RGB'][_LUA_RANDOMIZED_LINE]
      with self.assertRaises(
          AssertionError,
          msg=f'Episodes {episode} and {episode+1} match: {last_obs} == {obs}'):
        np.testing.assert_equal(last_obs, obs)

  def test_no_seed_causes_nondeterminism_for_lua(self):
    lab2d_settings = copy.deepcopy(_TEST_SETTINGS)
    lab2d_settings['simulation']['map'] = _LUA_RANDOMIZATION_MAP

    env1 = self.enter_context(builder.builder(lab2d_settings))
    env2 = self.enter_context(builder.builder(lab2d_settings))
    for episode in range(5):
      obs1 = env1.reset().observation['WORLD.RGB'][_LUA_RANDOMIZED_LINE]
      obs2 = env2.reset().observation['WORLD.RGB'][_LUA_RANDOMIZED_LINE]
      with self.assertRaises(
          AssertionError, msg=f'Episode {episode} match {obs1} == {obs2}'):
        np.testing.assert_equal(obs1, obs2)


if __name__ == '__main__':
  absltest.main()
